@@ -59,6 +59,7 @@ static const char *clutter_input_axis_atom_names[] = {
   "Abs Tilt X",         /* CLUTTER_INPUT_AXIS_XTILT */
   "Abs Tilt Y",         /* CLUTTER_INPUT_AXIS_YTILT */
   "Abs Wheel",          /* CLUTTER_INPUT_AXIS_WHEEL */
+  "Abs Distance",       /* CLUTTER_INPUT_AXIS_DISTANCE */
 };
 
 #define N_AXIS_ATOMS    G_N_ELEMENTS (clutter_input_axis_atom_names)
@@ -94,7 +95,7 @@ translate_valuator_class (Display             *xdisplay,
     }
 
   for (i = CLUTTER_INPUT_AXIS_IGNORE;
-       i <= CLUTTER_INPUT_AXIS_WHEEL;
+       i < CLUTTER_INPUT_AXIS_LAST;
        i += 1)
     {
       if (clutter_input_axis_atoms[i] == class->label)
@@ -732,12 +733,15 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
 
         device = g_hash_table_lookup (manager_xi2->devices_by_id,
                                       GINT_TO_POINTER (xev->deviceid));
-        _clutter_input_device_reset_axes (device);
-        _clutter_input_device_reset_scroll_info (device);
-        translate_device_classes (backend_x11->xdpy,
-                                  device,
-                                  xev->classes,
-                                  xev->num_classes);
+        if (device)
+          {
+            _clutter_input_device_reset_axes (device);
+            _clutter_input_device_reset_scroll_info (device);
+            translate_device_classes (backend_x11->xdpy,
+                                      device,
+                                      xev->classes,
+                                      xev->num_classes);
+          }
       }
       retval = CLUTTER_TRANSLATE_REMOVE;
       break;
@@ -828,12 +832,27 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
       {
         XIDeviceEvent *xev = (XIDeviceEvent *) xi_event;
 
+        source_device = g_hash_table_lookup (manager_xi2->devices_by_id,
+                                             GINT_TO_POINTER (xev->sourceid));
+        device = g_hash_table_lookup (manager_xi2->devices_by_id,
+                                      GINT_TO_POINTER (xev->deviceid));
+
+        /* Set the stage for core events coming out of nowhere (see bug #684509) */
+        if (clutter_input_device_get_device_mode (device) == CLUTTER_INPUT_MODE_MASTER &&
+            clutter_input_device_get_pointer_stage (device) == NULL &&
+            stage != NULL)
+          _clutter_input_device_set_stage (device, stage);
+
         switch (xev->detail)
           {
           case 4:
           case 5:
           case 6:
           case 7:
+            /* we only generate Scroll events on ButtonPress */
+            if (xi_event->evtype == XI_ButtonRelease)
+              return CLUTTER_TRANSLATE_REMOVE;
+
             event->scroll.type = event->type = CLUTTER_SCROLL;
 
             if (xev->detail == 4)
@@ -854,12 +873,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
               _clutter_input_device_xi2_translate_state (&xev->mods,
                                                          &xev->buttons);
 
-            source_device = g_hash_table_lookup (manager_xi2->devices_by_id,
-                                                 GINT_TO_POINTER (xev->sourceid));
             clutter_event_set_source_device (event, source_device);
-
-            device = g_hash_table_lookup (manager_xi2->devices_by_id,
-                                          GINT_TO_POINTER (xev->deviceid));
             clutter_event_set_device (event, device);
 
             event->scroll.axes = translate_axes (event->scroll.device,
@@ -868,10 +882,28 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
                                                  stage_x11,
                                                  &xev->valuators);
 
+            CLUTTER_NOTE (EVENT,
+                          "scroll: win:0x%x, device:%d '%s', time:%d "
+                          "(direction:%s, "
+                          "x:%.2f, y:%.2f, "
+                          "emulated:%s)",
+                          (unsigned int) stage_x11->xwin,
+                          device->id,
+                          device->device_name,
+                          event->any.time,
+                          event->scroll.direction == CLUTTER_SCROLL_UP ? "up" :
+                          event->scroll.direction == CLUTTER_SCROLL_DOWN ? "down" :
+                          event->scroll.direction == CLUTTER_SCROLL_LEFT ? "left" :
+                          event->scroll.direction == CLUTTER_SCROLL_RIGHT ? "right" :
+                          "invalid",
+                          event->scroll.x,
+                          event->scroll.y,
 #ifdef HAVE_XINPUT_2_2
-            if (xev->flags & XIPointerEmulated)
-              _clutter_event_set_pointer_emulated (event, TRUE);
-#endif /* HAVE_XINPUT_2_2 */
+                          (xev->flags & XIPointerEmulated) ? "yes" : "no"
+#else
+                          "no"
+#endif
+                          );
             break;
 
           default:
@@ -889,12 +921,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
               _clutter_input_device_xi2_translate_state (&xev->mods,
                                                          &xev->buttons);
 
-            source_device = g_hash_table_lookup (manager_xi2->devices_by_id,
-                                                 GINT_TO_POINTER (xev->sourceid));
             clutter_event_set_source_device (event, source_device);
-
-            device = g_hash_table_lookup (manager_xi2->devices_by_id,
-                                          GINT_TO_POINTER (xev->deviceid));
             clutter_event_set_device (event, device);
 
             event->button.axes = translate_axes (event->button.device,
@@ -902,24 +929,36 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
                                                  event->button.y,
                                                  stage_x11,
                                                  &xev->valuators);
+
+            CLUTTER_NOTE (EVENT,
+                          "%s: win:0x%x, device:%d '%s', time:%d "
+                          "(button:%d, "
+                          "x:%.2f, y:%.2f, "
+                          "axes:%s, "
+                          "emulated:%s)",
+                          event->any.type == CLUTTER_BUTTON_PRESS
+                            ? "button press  "
+                            : "button release",
+                          (unsigned int) stage_x11->xwin,
+                          device->id,
+                          device->device_name,
+                          event->any.time,
+                          event->button.button,
+                          event->button.x,
+                          event->button.y,
+                          event->button.axes != NULL ? "yes" : "no",
+#ifdef HAVE_XINPUT_2_2
+                          (xev->flags & XIPointerEmulated) ? "yes" : "no"
+#else
+                          "no"
+#endif
+                          );
             break;
           }
 
         if (source_device != NULL && device->stage != NULL)
           _clutter_input_device_set_stage (source_device, device->stage);
 
-        CLUTTER_NOTE (EVENT,
-                      "%s: win:0x%x, device:%s (button:%d, x:%.2f, y:%.2f, axes:%s)",
-                      event->any.type == CLUTTER_BUTTON_PRESS
-                        ? "button press  "
-                        : "button release",
-                      (unsigned int) stage_x11->xwin,
-                      event->button.device->device_name,
-                      event->button.button,
-                      event->button.x,
-                      event->button.y,
-                      event->button.axes != NULL ? "yes" : "no");
-
 #ifdef HAVE_XINPUT_2_2
         if (xev->flags & XIPointerEmulated)
           _clutter_event_set_pointer_emulated (event, TRUE);
@@ -939,6 +978,14 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
 
         source_device = g_hash_table_lookup (manager_xi2->devices_by_id,
                                              GINT_TO_POINTER (xev->sourceid));
+        device = g_hash_table_lookup (manager_xi2->devices_by_id,
+                                      GINT_TO_POINTER (xev->deviceid));
+
+        /* Set the stage for core events coming out of nowhere (see bug #684509) */
+        if (clutter_input_device_get_device_mode (device) == CLUTTER_INPUT_MODE_MASTER &&
+            clutter_input_device_get_pointer_stage (device) == NULL &&
+            stage != NULL)
+          _clutter_input_device_set_stage (device, stage);
 
         if (scroll_valuators_changed (source_device,
                                       &xev->valuators,
@@ -957,14 +1004,12 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
 
             clutter_event_set_scroll_delta (event, delta_x, delta_y);
             clutter_event_set_source_device (event, source_device);
-
-            device = g_hash_table_lookup (manager_xi2->devices_by_id,
-                                          GINT_TO_POINTER (xev->deviceid));
             clutter_event_set_device (event, device);
 
             CLUTTER_NOTE (EVENT,
-                          "smooth scroll: win:0x%x device:%s (x:%.2f, y:%.2f, delta:%f, %f)",
+                          "smooth scroll: win:0x%x device:%d '%s' (x:%.2f, y:%.2f, delta:%f, %f)",
                           (unsigned int) stage_x11->xwin,
+                          event->scroll.device->id,
                           event->scroll.device->device_name,
                           event->scroll.x,
                           event->scroll.y,
@@ -986,9 +1031,6 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
                                                      &xev->buttons);
 
         clutter_event_set_source_device (event, source_device);
-
-        device = g_hash_table_lookup (manager_xi2->devices_by_id,
-                                      GINT_TO_POINTER (xev->deviceid));
         clutter_event_set_device (event, device);
 
         event->motion.axes = translate_axes (event->motion.device,
@@ -1005,8 +1047,9 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
           _clutter_event_set_pointer_emulated (event, TRUE);
 #endif /* HAVE_XINPUT_2_2 */
 
-        CLUTTER_NOTE (EVENT, "motion: win:0x%x device:%s (x:%.2f, y:%.2f, axes:%s)",
+        CLUTTER_NOTE (EVENT, "motion: win:0x%x device:%d '%s' (x:%.2f, y:%.2f, axes:%s)",
                       (unsigned int) stage_x11->xwin,
+                      event->motion.device->id,
                       event->motion.device->device_name,
                       event->motion.x,
                       event->motion.y,
@@ -1050,9 +1093,6 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
                                             stage_x11,
                                             &xev->valuators);
 
-        if (source_device != NULL && device->stage != NULL)
-          _clutter_input_device_set_stage (source_device, device->stage);
-
         if (xi_event->evtype == XI_TouchBegin)
           {
             event->touch.modifier_state |= CLUTTER_BUTTON1_MASK;
@@ -1065,10 +1105,12 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
         if (xev->flags & XITouchEmulatingPointer)
           _clutter_event_set_pointer_emulated (event, TRUE);
 
-        CLUTTER_NOTE (EVENT, "touch %s: win:0x%x device:%s (x:%.2f, y:%.2f, axes:%s)",
+        CLUTTER_NOTE (EVENT, "touch %s: win:0x%x device:%d '%s' (seq:%d, x:%.2f, y:%.2f, axes:%s)",
                       event->type == CLUTTER_TOUCH_BEGIN ? "begin" : "end",
                       (unsigned int) stage_x11->xwin,
+                      event->touch.device->id,
                       event->touch.device->device_name,
+                      GPOINTER_TO_UINT (event->touch.sequence),
                       event->touch.x,
                       event->touch.y,
                       event->touch.axes != NULL ? "yes" : "no");
@@ -1103,9 +1145,6 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
                                             stage_x11,
                                             &xev->valuators);
 
-        if (source_device != NULL && device->stage != NULL)
-          _clutter_input_device_set_stage (source_device, device->stage);
-
         event->touch.modifier_state =
           _clutter_input_device_xi2_translate_state (&xev->mods,
                                                      &xev->buttons);
@@ -1114,9 +1153,11 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
         if (xev->flags & XITouchEmulatingPointer)
           _clutter_event_set_pointer_emulated (event, TRUE);
 
-        CLUTTER_NOTE (EVENT, "touch update: win:0x%x device:%s (x:%.2f, y:%.2f, axes:%s)",
+        CLUTTER_NOTE (EVENT, "touch update: win:0x%x device:%d '%s' (seq:%d, x:%.2f, y:%.2f, axes:%s)",
                       (unsigned int) stage_x11->xwin,
+                      event->touch.device->id,
                       event->touch.device->device_name,
+                      GPOINTER_TO_UINT (event->touch.sequence),
                       event->touch.x,
                       event->touch.y,
                       event->touch.axes != NULL ? "yes" : "no");
@@ -1149,7 +1190,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
             event->crossing.x = xev->event_x;
             event->crossing.y = xev->event_y;
 
-            _clutter_stage_add_device (stage, device);
+            _clutter_input_device_set_stage (device, stage);
           }
         else
           {
@@ -1173,7 +1214,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
             event->crossing.x = xev->event_x;
             event->crossing.y = xev->event_y;
 
-            _clutter_stage_remove_device (stage, device);
+            _clutter_input_device_set_stage (device, NULL);
           }
 
         _clutter_input_device_reset_scroll_info (source_device);
